
#ifndef PLS_ALIGNED_STACK_H
#define PLS_ALIGNED_STACK_H

#include <cstdint>
#include <cstdlib>

#include "pls/internal/base/error_handling.h"
#include "pls/internal/base/alignment.h"

namespace pls {
    namespace internal {
        namespace data_structures {
            /**
             * Generic stack-like data structure that allows to allocate arbitrary objects in a given memory region.
             * The objects will be stored aligned in the stack, making the storage cache friendly and very fast
             * (as long as one can live with the stack restrictions).
             *
             * IMPORTANT: Does not call destructors on stored objects! Do not allocate resources in the objects!
             *
             * Usage:
             * aligned_stack stack{pointer_to_memory, size_of_memory};
             * T* pointer = stack.push(some_object); // Copy-Constrict the object on top of stack
             * stack.pop<T>(); // Deconstruct the top object of type T
             */
            class aligned_stack {
                // Keep bounds of our memory block
                char* memory_start_;
                char* memory_end_;

                // Current head will always be aligned to cache lines
                char* head_;
            public:
                typedef char* state;

                aligned_stack(): memory_start_{nullptr}, memory_end_{nullptr}, head_{nullptr} {};
                aligned_stack(char* memory_region, std::size_t size);

                template<typename T>
                T* push(const T& object) {
                    // Copy-Construct
                    return new ((void*)push<T>())T(object);
                }

                template<typename T>
                void* push() {
                    void* result = reinterpret_cast<T*>(head_);

                    // Move head to next aligned position after new object
                    head_ = base::alignment::next_alignment(head_ + sizeof(T));
                    if (head_ >= memory_end_) {
                        PLS_ERROR("Tried to allocate object on alligned_stack without sufficient memory!");
                    }

                    return result;
                }

                template<typename T>
                T pop() {
                    head_ = head_ - base::alignment::next_alignment(sizeof(T));
                    return *reinterpret_cast<T*>(head_);
                }

                state save_state() {
                    return head_;
                }

                void reset_state(state new_state) {
                    head_ = new_state;
                }
            };
        }
    }
}

#endif //PLS_ALIGNED_STACK_H
